/**
 * Copyright © 2014-2015 Paolo Simonetto
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package ocotillo.graph.serialization.oco;

import ocotillo.graph.Attribute;
import ocotillo.graph.AttributeType;
import ocotillo.graph.Edge;
import ocotillo.graph.EdgeAttribute;
import ocotillo.graph.Element;
import ocotillo.graph.Graph;
import ocotillo.graph.Node;
import ocotillo.graph.NodeAttribute;
import ocotillo.graph.StdAttribute;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Writer for oco files.
 */
public class OcoWriter {

    private final OcoConverterSet converters;

    /**
     * Constructs an oco writer.
     *
     * @param converters the converters.
     */
    protected OcoWriter(OcoConverterSet converters) {
        this.converters = converters;
    }

    /**
     * Writes a graph in the oco format.
     *
     * @param graph the input graph.
     * @return the graph description in oco format.
     */
    protected List<String> write(Graph graph) {
        List<String> lines = new LinkedList<>();
        writeGraph(graph, 0, lines);
        return lines;
    }

    /**
     * Recursively write graphs in oco format.
     *
     * @param graph the graph.
     * @param graphLevel the current graph level.
     * @param lines the file lines.
     */
    private void writeGraph(Graph graph, int graphLevel, List<String> lines) {
        writeGraphHeader(graphLevel, lines);
        writeAttributesAndElements(graph, AttributeType.graph, lines);
        lines.add("");
        lines.add("#nodes");
        writeAttributesAndElements(graph, AttributeType.node, lines);
        lines.add("");
        lines.add("#edges");
        writeAttributesAndElements(graph, AttributeType.edge, lines);
        lines.add("");
        lines.add("");

        for (Graph subgraph : graph.subGraphs()) {
            writeGraph(subgraph, graphLevel + 1, lines);
        }
    }

    /**
     * Writes the graph header according to the current graph level.
     *
     * @param graphLevel the current graph level.
     * @param lines the file lines.
     */
    private void writeGraphHeader(int graphLevel, List<String> lines) {
        StringBuilder headerPrefix = new StringBuilder("#");
        for (int i = 0; i < graphLevel; i++) {
            headerPrefix.append("#");
        }
        lines.add(headerPrefix.append("graph").toString());
    }

    /**
     * Write the attribute and the eventual elements of a oco file block.
     *
     * @param graph the graph.
     * @param generalAttrType the kind of block to written for the graph.
     * @param lines the file lines.
     */
    @SuppressWarnings({"unchecked", "rawtypes"})
    private void writeAttributesAndElements(Graph graph, AttributeType generalAttrType, List<String> lines) {
        StringBuilder attributes = new StringBuilder("@attribute\t");
        StringBuilder types = new StringBuilder("@type\t");
        StringBuilder defaults = new StringBuilder("@default\t");

        boolean isRootGraph = (graph == graph.rootGraph());
        if (generalAttrType == AttributeType.edge && isRootGraph) {
            attributes.append("@from\t@to\t");
            types.append("\t\t");
            defaults.append("\t\t");
        }

        Map<Element, StringBuilder> elements = prepareElementsMap(graph, generalAttrType, isRootGraph);

        for (String attributeName : graph.localAttributes(generalAttrType).keySet()) {
            Attribute attribute = graph.attribute(generalAttrType, attributeName);
            if (!attribute.isSleeping()) {
                Object defaultValue = attribute.getDefault();
                OcoValueConverter converter = getConverter(attributeName, defaultValue);
                String attributeType = converter.typeName();
                String attributeDefault = converter.graphLibToOco(defaultValue);

                attributes.append(attributeName).append("\t");
                types.append(attributeType).append("\t");
                defaults.append(attributeDefault).append("\t");

                if (generalAttrType == AttributeType.node) {
                    for (Node node : graph.nodes()) {
                        NodeAttribute nodeAttribute = (NodeAttribute) attribute;
                        String value = converter.graphLibToOco(nodeAttribute.get(node)) + "\t";
                        elements.get(node).append(value);
                    }
                } else if (generalAttrType == AttributeType.edge) {
                    for (Edge edge : graph.edges()) {
                        EdgeAttribute edgeAttribute = (EdgeAttribute) attribute;
                        String value = converter.graphLibToOco(edgeAttribute.get(edge)) + "\t";
                        elements.get(edge).append(value);
                    }
                }
            }
        }
        lines.add(attributes.substring(0, attributes.length() - 1));
        lines.add(types.substring(0, types.length() - 1));
        lines.add(defaults.substring(0, defaults.length() - 1));
        writeElements(elements, lines);
    }

    /**
     * Gets a converter for the given attribute.
     *
     * @param attributeName the attribute name.
     * @param defaultValue the attribute default value.
     * @return the converter.
     */
    private OcoValueConverter<?> getConverter(String attributeName, Object defaultValue) {
        if (StdAttribute.isStandard(attributeName)) {
            Class<?> type = StdAttribute.get(attributeName).matchingClass;
            return converters.get(type);
        } else {
            Class<?> type = defaultValue.getClass();
            if (converters.contains(type)) {
                return converters.get(type);
            } else {
                return converters.get(String.class);
            }
        }
    }

    /**
     * Prepares the map of elements for the block.
     *
     * @param graph the graph.
     * @param generalAttrType the kind of block to written for the graph.
     * @param isRootGraph indicates if the graph is the root one.
     * @return the element map.
     */
    private Map<Element, StringBuilder> prepareElementsMap(Graph graph, AttributeType generalAttrType, boolean isRootGraph) {
        Map<Element, StringBuilder> elements = new HashMap<>();
        if (generalAttrType == AttributeType.node) {
            for (Node node : graph.nodes()) {
                elements.put(node, new StringBuilder(node.id() + "\t"));
            }
        } else if (generalAttrType == AttributeType.edge) {
            for (Edge edge : graph.edges()) {
                StringBuilder edgeLine = new StringBuilder(edge.id() + "\t");
                if (isRootGraph) {
                    edgeLine.append(edge.source().id()).append("\t")
                            .append(edge.target().id()).append("\t");
                }
                elements.put(edge, edgeLine);
            }
        }
        return elements;
    }

    /**
     * Writes the elements in order by id.
     *
     * @param elements the element map.
     * @param lines the file lines.
     */
    private void writeElements(Map<Element, StringBuilder> elements, List<String> lines) {
        List<Element> elementList = new ArrayList<>(elements.keySet());
        Collections.sort(elementList, new Comparator<Element>() {

            @Override
            public int compare(Element a, Element b) {
                return a.id().compareTo(b.id());
            }
        });

        for (Element element : elementList) {
            StringBuilder elementLine = elements.get(element);
            lines.add(elementLine.substring(0, elementLine.length() - 1));
        }
    }
}
